 /*
  * proc.c
  * Copyright (C) Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
  * 
  * simple-netaid is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
  * Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  * 
  * simple-netaid is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  * See the GNU General Public License for more details.
  * 
  * You should have received a copy of the GNU General Public License along
  * with this program.  If not, see <http://www.gnu.org/licenses/>.
  * 
  * See the COPYING file.
  */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <stdbool.h>
#include <ctype.h>       /*    */

#include <simple-netaid/sbuf.h>

#include "def.h"
#include "config.h"
#include "proc.h"

#define SKIP(p) while (*p && isspace (*p)) { p++; }
#define WANT(p) *p && ! isspace (*p)
#define OK 0
#define MALLOC_FAILED -1
#define MISCOUNT -2
#define NO_INPUT -3

#define MAX_SIZE 1024


static
int count_args (const char * input)
{
    const char * p;
    int argc = 0;
    p = input;
    while (*p) {
        SKIP (p);
        if (WANT (p)) {
            argc++;
            while (WANT (p)) {
                p++;
       }
        }
    }
    return argc;
}

static
void free_argv (char ** argv, int i)
{
    int j;
    for (j = 0; j < i; j++) {
   free (argv[j]);
    }
}

static
int copy_args (const char * input, int argc, char ** argv)
{
    int i = 0;
    const char *p;

    p = input;
    while (*p) {
        SKIP (p);
        if (WANT (p)) {
            const char * end = p;
            char * copy;
            while (WANT (end)) {
                end++;
       }
            copy = argv[i] = malloc (end - p + 1);
            if (! argv[i]) {
      free_argv (argv, i);
                return MALLOC_FAILED;
       }
            while (WANT (p)) {
                *copy = *p;
      copy++;
      p++;
       }
            *copy = '\0';
            i++;
        }
    }
    if (i != argc) {
   free_argv (argv, i);
   return MISCOUNT;
    }
    return OK;
}

#undef SKIP
#undef WANT

int argc_argv (const char * input, int * argc_ptr, char *** argv_ptr)
{
    int argc;
    char ** argv;
    int status;

    argc = count_args (input);
    if (argc == 0) {
        return NO_INPUT;
    }
    argv = malloc (sizeof (char *) * argc);
    if (! argv) {
        return MALLOC_FAILED;
    }
    status = copy_args (input, argc, argv);

    switch (status) {
       case OK:
         *argc_ptr = argc;
         *argv_ptr = argv;
         return OK;
   
       case MALLOC_FAILED:
         free (argv);
         return MALLOC_FAILED;
   
       default:
         fprintf (stderr, "copy_args failed with status %d for '%s'.\n", status, input);
         exit (EXIT_FAILURE);
    }
}

int termproc(int interval)
{   
   int count = 0;
   FILE *fp = NULL;
   char cmd[16] = {0};
   char output[MAX_SIZE]={0};
        
   sprintf(cmd, "pidof snetaid");
   fp=popen(cmd, "r");
   if(!fp) {
      log_error("%s: error opening the pipe %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE);
   }
         
   if(fgets (output, MAX_SIZE, fp)) {
      int argc;
      char ** argv;   
      if(*output != 0) { 
         output[strcspn(output, "\n")] = 0;
         if (argc_argv (output, & argc, & argv) != OK) {
            log_error("%s: Error splitting pids %s\n", progname, strerror(errno));
            exit(EXIT_FAILURE);
         }
    
         for(int i = 0; i < argc; i++) { 
            int mismatch = -1; 
            bool res = false;
            res = request_info(&mismatch, argv[i]);
            if(res) {
               count++;
               int cpid = atoi(argv[i]);
               if(cpid != (uint32_t)getpid()) {
                  if( interval == 0 || ((mismatch>-1) && (mismatch<interval)) ) { 
                     kill(cpid, SIGTERM);
                  }
               }
            }
         }
      }
   }
   
   if(fp) pclose(fp);
       
   return count;
}

bool request_info(int *mismatch, const char *cpid)
{
   struct sbuf s __cleanbuf__;   
   char haystack[MAX_SIZE] = {0};
   char cmd[128] = {0};
   FILE *pf = NULL;
   bool res = false;
   int start_time = 0, m_start_time = 0;
   struct sbuf m_buf;
   pid_t pid;
   char *mypid = NULL;
   
   pid = getpid();     
   if(asprintf(&mypid, "%jd", (intmax_t)pid) != -1) {
      sbuf_init(&m_buf);  
      get_pstat_info(&m_buf, &m_start_time, mypid);
   } else {
      log_error("%s: Error getting info %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE);
   }
     
   sbuf_init(&s);
   get_pstat_info(&s, &start_time, cpid);   

   if(s.len == 0) {     /* FIXME */
	  *mismatch = 0;
	  return true;
   }
   
   if(m_start_time && start_time && (start_time >= m_start_time)) {
      *mismatch = start_time - m_start_time;
   }
   
   // Do they match?
   // s.buf -> path to the binary of the process to be inspected
   // m_buf.buf -> path to the binary of the running process 
   //              gotten from getpid() via pstat
   // NOTE: scripts like '/etc/init.d/snetaid' will return "/bin/dash", 
   //       "/bin/bash", and so on.
   if(strcmp(s.buf, m_buf.buf)) {
      sbuf_free(&s);
      sbuf_free(&m_buf);
      return false;
   } 

   sbuf_free(&m_buf);
   
   // They match, however we won't blithely kill the process
   // Make sure it depends on libnetaid, before killing it
   // NOTE: I agree this migth be redundant, because we already 
   //       have compared both paths. But go ahead...
   sprintf(cmd, "ldd ");
   strcat(cmd, s.buf);
   strcat(cmd, " | grep libnetaid.so");   
   pf = popen(cmd, "r");
   if(!pf) {  
      sbuf_free(&s);
      log_error("%s: popen(): %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE);
   }   
   if(fgets(haystack, MAX_SIZE, pf)) {
      // The following verification is recommended because ldd might return a 
      // warning. Hence, we would get a non empty output.
      const char *needle = "libnetaid.so";
      if(strstr(haystack, needle)) res = true;
   } 
   
   sbuf_free(&s);   
   pclose(pf);
      
   return res;
}

void get_pstat_info(struct sbuf *s, int *start_time, const char *cpid)
{
   char output[MAX_SIZE] = {0};
   struct sbuf cmd;
   FILE *fp = NULL;
  
   // Get the output of "pstat PID" and inspect the PSTAT_BINARY="value" key pair
   sbuf_init(&cmd);
   sbuf_concat(&cmd, 2, "/usr/bin/pstat ", cpid);
  
   fp = popen(cmd.buf, "r");
   if(!fp) {
      log_error("%s: popen(): %s\n", progname, strerror(errno));
      free(cmd.buf);
      exit(EXIT_FAILURE);
   }
   sbuf_free(&cmd);
   
   while(fgets(output, MAX_SIZE, fp)) {
     const char delim[2] = "=";
     char *token = strtok( (char*)output, delim );
     if(!strcmp(token, "PSTAT_BINARY")) {
      int rc __attribute__((unused)) = 0;
        token = strtok(NULL, delim);
        token[strcspn(token, "\n")] = 0;
        rc = strdelch(token, '"');
        sbuf_addstr(s, token);
     }
     if(start_time && !strcmp(token, "PSTAT_STARTTIME")) {
      int rc __attribute__((unused)) = 0;
        token = strtok(NULL, delim);
        token[strcspn(token, "\n")] = 0;
        *start_time=atoi(token);
     }
   }
        
   pclose(fp);
}
   

